import os
from pathlib import Path
import glob
import shutil
import argparse
import numpy as np
import pandas as pd

from rdfpy import rdf, rdf3d


class rdfpy_tools():

    def __init__(self, moldict):
        self.moldict = moldict


    def get_rdf_single(self, atoms, atom_type1, atom_type2):

    def get_rdf(self, atom_type1, atom_type2):
        rdf={}
        all_at_coords  = []
        for k, m in self.moldict.items():
            # rdfpy works on coordinates passed as a numpy array,
            # so we prepare one that collects coordinates
            # of two selected atom types
            at  = m.GetAtoms()
            at_symbols = [a.GetSymbol() for a in at]
            for i, a in enumerate(at):
                for at1_ind in atom_type1[k]:
                    if at1_ind == i:
                        coords = m.GetConformer().GetAtomPosition(i)
                        all_at_coords.append([coords.x, coords.y, coords.z])
                for at2_ind in atom_type2[k]:
                    if at2_ind == i:
                        coords = m.GetConformer().GetAtomPosition(i)
                        all_at_coords.append([coords.x, coords.y, coords.z])
                #print('TEST k, at c1.x, c1.y, c1.z ', k, at_symbols[i], coords.x, coords.y, coords.z)
            
        coords_array = np.array(all_at_coords)
        g_r, radii = rdf3d(coords_array, dr=0.1)
        return coords_array, g_r, radii 


class dataout():

    """
    This class collects examples of how to save data on files
    """

    def __init__(self, data_dir=None):
        self.data_dir = data_dir


    def data_dict_to_dataframe(self, data_dict):
        """
        entering data_dict is a dict of dict
        """
        result = []
        for p, d in data_dict.items():
            for m, v in d.items():
                for s in v:
                    data = {'parameter':p, 'mol':m, 'value':s}
                    result.append(data)
        data_array = pd.DataFrame(result)
        return data_array

    def write_to_csvfile(self, data_filename=None, dataframe=None):
        if data_filename is not None:
            Path(self.data_dir).mkdir(parents=True, exist_ok=True)
            f=Path(self.data_dir, data_filename).resolve()
            if dataframe is not None:
                dataframe.to_csv(f, index=False)


class plots():

    """
    This class collects examples of plots done with matplotlib
    """

    def __init__(self, plot_dir=None):
        self.plot_dir = plot_dir


    def plot_atoms_as_points(self, g_r, radii, coords=None, plot_filename=None, plot_title=None):
        if coords is not None:
            fig = plt.figure(figsize=(9, 9))
            ax = fig.add_subplot(111, projection='3d')
            ax.scatter(coords[:, 0], coords[:, 1], coords[:, 2], color='aqua', alpha=0.3, edgecolors='k', s=25)
            plt.axis('off')
            if plot_filename is not None:
                Path(self.plot_dir).mkdir(parents=True, exist_ok=True)
                f=Path(self.plot_dir, plot_filename).resolve()
                plt.savefig(f, bbox_inches='tight', pad_inches=0.0)
                plt.close()
            else:
                return plt


    def plot_rdf_and_atoms(self, g_r, radii, coords=None, plot_filename=None, plot_title=None):
        if coords is not None:
            tempfile='atoms.png'
            self.plot_atoms_as_points(g_r, radii, coords=coords, plot_filename=tempfile)
            f = Path(self.plot_dir, tempfile).resolve()
            image = plt.imread(f)

            fig, axes = plt.subplots(1, 2, figsize=(12, 4), gridspec_kw={'width_ratios': [1, 2]})

            axes[0].imshow(image)
            axes[0].axis('off')

            axes[1].plot(radii, g_r, color='k', alpha=0.75)
            axes[1].hlines(y=1.0, xmin=0.0, xmax=max(radii), color='r', linestyle='--', alpha=0.4)
            axes[1].set_ylabel(r'g(r)')
            axes[1].set_xlabel(r'r')
            axes[1].set_xlim(0.0, max(radii))
            if plot_title is not None:
                axes[0].set_title(plot_title)
            if plot_filename is not None:
                Path(self.plot_dir).mkdir(parents=True, exist_ok=True)
                f = Path(self.plot_dir, plot_filename).resolve()
                plt.savefig(f, bbox_inches='tight', pad_inches=0.0)
                plt.close()
            else:
                plt.show()
                return plt



class crest_analysis():
    def __init__(self, crest_parentdirname, qm_calcdirname, qm_tpldirname):
        self.crest_parentdirname = crest_parentdirname
        self.qm_parentdirname    = qm_calcdirname
        self.qm_templatedirname  = qm_tpldirname

    def ignore_files(self, dirname, files):
        return [f for f in files if os.path.isfile(os.path.join(dirname, f))]

    def setup_qm_dirstructure(self, dir_template=None, dir_result=None):
        if dir_template is None:
            dir_template = Path(self.crest_parentdirname)
        if dir_result is None:
            dir_result = self.qm_parentdirname
            Path(dir_result).mkdir(parents=True, exist_ok=True)
        shutil.copytree(dir_template, dir_result, ignore=self.ignore_files, dirs_exist_ok=True)

    def get_single_struc(self, finp):
        data = {}
        with open(finp, 'r') as f:
            lines = f.readlines()
            nr_struct = 0
            new_struct_line = 0
            for i, line in enumerate(lines):
                if new_struct_line < len(lines):
                    nr_at = int(lines[new_struct_line])
                    energy = float(lines[new_struct_line+1])
                    coords = [x.strip() for x in lines[new_struct_line+2:new_struct_line+2+nr_at]]
                    nr_struct += 1
                    new_struct_line = new_struct_line + (2 + nr_at)*nr_struct
                    data[nr_struct] = [str(nr_at)] + [str(energy)] + [x for x in coords]
        return data


    def parse_structures(self, input_info=None, output_dir=None):
        files = Path(self.crest_parentdirname).rglob("*.xyz")
        output_files_list = []
        if output_dir is not None:
            qm_parent = output_dir[0]
            qm_coords = output_dir[1]
        else:
            qm_parent = self.qm_parentdirname
            qm_coords = 'coordinates'
        for f in files:
            if f.name == input_info[1]:
                if input_info[0] in f.parts:
                    # TODO this is error prone, be careful
                    molgroup_name = f.parts[1]
                    data = self.get_single_struc(str(f.resolve()))
                    for k, v in data.items():
                        mol_name = 'struc_'+str(k)
                        fileout_name = molgroup_name + '_' + mol_name + '.xyz'
                        fileout_dir  = Path('/'.join([x.replace(self.crest_parentdirname, qm_parent) for x in f.parts[:-1]]), qm_coords)
                        fileout_dir.mkdir(parents=True, exist_ok=True)
                        fileout = Path(fileout_dir, fileout_name).resolve()
                        output_files_list.append(fileout)
                        with open(str(fileout), 'w') as fout:
                            for l in v:
                                fout.write('{}\n'.format(l))
        return output_files_list
                
    def prep_for_pyadf(self, inpfiles, tpl_files):
        for inps in inpfiles:
            calcdir = inps.resolve().parents[1]
            print(calcdir)
            for tpl in tpl_files:
                shutil.copy(Path(self.qm_templatedirname, tpl).resolve(), calcdir)


